//
// LDASTools utilities - A collection of utilities base on LDASTools Suite
//
// Copyright (C) 2023 California Institute of Technology
//
// LDASTools utilities is free software; you may redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 (GPLv2) of the
// License or at your discretion, any later version.
//
// LDASTools utilities is distributed in the hope that it will be useful, but
// without any warranty or even the implied warranty of merchantability
// or fitness for a particular purpose. See the GNU General Public
// License (GPLv2) for more details.
//
// Neither the names of the California Institute of Technology (Caltech),
// The Massachusetts Institute of Technology (M.I.T), The Laser
// Interferometer Gravitational-Wave Observatory (LIGO), nor the names
// of its contributors may be used to endorse or promote products derived
// from this software without specific prior written permission.
//
// You should have received a copy of the licensing terms for this
// software included in the file LICENSE located in the top-level
// directory of this package. If you did not, you can view a copy at
// http://dcc.ligo.org/M1500244/LICENSE
//

#include "genericAPI/TCL.hh"

#include "query.hh"

namespace LDAS
{
    namespace createRDS
    {
        //=======================================================================
        //
        //=======================================================================
        query::query( const std::string Query ) : ext( ".gwf" )
        {
            static const char* caller = "query::query";

            std::list< std::string >   parameters;
            std::vector< std::string > op;
            int                        count = 0;

            GenericAPI::TCL::ParseList( Query, parameters );
            op.resize( std::min( parameters.size( ),
                                 size_t( POSITION_QUERY_LIST_SIZE_TCL ) ) );
            QUEUE_LOG_MESSAGE( "op.size: " << op.size( ),
                               MT_DEBUG,
                               30,
                               caller,
                               "CMD_CREATE_RDS" );
            for ( std::list< std::string >::const_iterator
                      cur = parameters.begin( ),
                      last = parameters.end( );
                  cur != last;
                  ++cur, ++count )
            {
                if ( count > POSITION_CHANNEL_LIST_TCL )
                {
                    op[ POSITION_CHANNEL_LIST_TCL ] += *cur;
                }
                else
                {
                    op[ count ] = *cur;
                }
            }
            QUEUE_LOG_MESSAGE(
                "number of elements in query: " << parameters.size( ),
                MT_DEBUG,
                30,
                caller,
                "CMD_CREATE_RDS" );
            desc = op[ POSITION_DESC_TCL ];
            ifo = op[ POSITION_IFO_TCL ];

            //---------------------------------------------------------------------
            // Parse the time range
            //---------------------------------------------------------------------
            std::string s( op[ POSITION_TIME_RANGE_TCL ] );
            std::string e;
            size_t      offset( s.find( '-' ) );

            if ( offset != std::string::npos )
            {
                e = s.substr( offset + 1 );
                s = s.substr( 0, offset );
            }

            LDASTools::AL::GPSTime::seconds_type sv;
            LDASTools::AL::GPSTime::seconds_type ev;

            {
                std::istringstream ss( s );
                ss >> sv;
            }
            {
                std::istringstream ss( e );
                ss >> ev;
            }

            start = LDASTools::AL::GPSTime( sv, 0 );
            end = LDASTools::AL::GPSTime( ev, 0 );

            //---------------------------------------------------------------------
            // Parse out the channel list
            //---------------------------------------------------------------------
            const std::string& cl( op[ POSITION_CHANNEL_LIST_TCL ] );
            size_t             clb( cl.find( '(' ) );
            size_t             cle( cl.rfind( ')' ) );

            if ( clb != std::string::npos )
            {
                ++clb;
            }
            QUEUE_LOG_MESSAGE(
                "cl: " << cl << " clb: " << clb << " cle: " << cle
                       << " len: " << ( cle - clb )
                       << " substr: " << cl.substr( clb, ( cle - clb ) ),
                MT_DEBUG,
                30,
                caller,
                "CMD_CREATE_RDS" );
            std::string cls( cl.substr( clb, ( cle - clb ) ) );
            //---------------------------------------------------------------------
            // Checking to see if the buffer names a file
            //---------------------------------------------------------------------
            {
                QUEUE_LOG_MESSAGE( "Checking if string is a filename",
                                   MT_DEBUG,
                                   30,
                                   caller,
                                   "CMD_CREATE_RDS" );
                std::regex filename_expr( "^[[:space:]]*(.*)[[:space:]]*$",
                                          std::regex::extended );
                std::regex channel_expr(
                    "^[[:space:]]*([^[:space:]]+)[[:space:]]+"
                    "([[:digit:]]+)[[:space:]]*$",
                    std::regex::extended );
                std::smatch        what;
                std::ostringstream channels;

                QUEUE_LOG_MESSAGE( "Checking if string is a filename: " << cls,
                                   MT_DEBUG,
                                   30,
                                   caller,
                                   "CMD_CREATE_RDS" );

                if ( std::regex_search( cls, what, filename_expr ) )
                {
                    std::ifstream s( what.str( 0 ).c_str( ) );

                    if ( s.is_open( ) )
                    {
                        QUEUE_LOG_MESSAGE(
                            "Channel name coming from file: " << what[ 0 ],
                            MT_DEBUG,
                            30,
                            caller,
                            "CMD_CREATE_RDS" );
                        try
                        {
                            bool        first = true;
                            std::smatch line_match;

                            while ( s.good( ) )
                            {
                                std::string line;

                                std::getline( s, line );
                                QUEUE_LOG_MESSAGE(
                                    "Parsing channel line: " << line,
                                    MT_DEBUG,
                                    30,
                                    caller,
                                    "CMD_CREATE_RDS" );
                                if ( std::regex_search(
                                         line, line_match, channel_expr ) )
                                {
                                    QUEUE_LOG_MESSAGE(
                                        "Found Matches: " << line_match.size( ),
                                        MT_DEBUG,
                                        30,
                                        caller,
                                        "CMD_CREATE_RDS" );
                                    if ( first )
                                    {
                                        first = false;
                                    }
                                    else
                                    {
                                        channels << ",";
                                    }

                                    switch ( line_match.size( ) )
                                    {
                                    case 1:
                                        channels << line_match[ 0 ];
                                        break;
                                    case 2:
                                        channels << line_match[ 0 ] << "!"
                                                 << line_match[ 1 ];
                                        break;
                                    }
                                }
                            }
                        }
                        catch ( ... )
                        {
                        }
                        s.close( );
                        cls = channels.str( );
                    }
                    else
                    {
                        QUEUE_LOG_MESSAGE( "Channel name is: " << what[ 0 ],
                                           MT_DEBUG,
                                           30,
                                           caller,
                                           "CMD_CREATE_RDS" );

                        cls = what[ 0 ];
                    }
                    QUEUE_LOG_MESSAGE( "Channels: " << cls,
                                       MT_DEBUG,
                                       30,
                                       caller,
                                       "CMD_CREATE_RDS" );
                }
            }

            offset = 0;

            size_t offset_end = cls.find( ',', offset );

            while ( offset != std::string::npos )
            {
                channels.push_back(
                    cls.substr( offset,
                                ( offset_end == std::string::npos )
                                    ? offset_end
                                    : ( offset_end - offset ) ) );
                QUEUE_LOG_MESSAGE( "Added channel: " << channels.back( ),
                                   MT_DEBUG,
                                   30,
                                   caller,
                                   "CMD_CREATE_RDS" );
                offset = offset_end;
                if ( offset != std::string::npos )
                {
                    ++offset;
                    offset_end = cls.find( ',', offset );
                }
            }
            QUEUE_LOG_MESSAGE(
                "finished query", MT_DEBUG, 30, caller, "CMD_CREATE_RDS" );
        }
    } // namespace createRDS
} // namespace LDAS
